""" This module provides the CHAODA algorithms implemented on top of CLAM.
"""
import logging
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import numpy as np

from pyclam.criterion import Criterion
from pyclam.criterion import MaxDepth
from pyclam.criterion import MetaMLSelect
from pyclam.criterion import MinPoints
from pyclam.criterion import PropertyThreshold
from pyclam.manifold import Cluster
from pyclam.manifold import Graph
from pyclam.manifold import Manifold
from pyclam.types import Metric
from pyclam.utils import catch_normalization_mode
from pyclam.utils import normalize

ClusterScores = Dict[Cluster, float]
Scores = Dict[int, float]
VOTING_MODES = [
    'mean',
    'product',
    'median',
    'min',
    'max',
    'p25',
    'p75',
]


class CHAODA:
    method_names = [
        'cluster_cardinality',
        'component_cardinality',
        'graph_neighborhood',
        'parent_cardinality',
        'stationary_probabilities',
        'vertex_degree',
    ]
    """ This class provides implementations of the CHAODA algorithms on top of the CLAM framework.
    """
    # TODO: Allow weights for each method in ensemble.
    # TODO: Allow meta-ml with users' own datasets.
    # TODO: Look at DSD (diffusion-state-distance) as a possible individual-method.
    # noinspection PyTypeChecker
    def __init__(
            self, *,
            metrics: Optional[List[Metric]] = None,
            min_depth: int = 4,
            max_depth: int = 25,
            min_points: int = 1,
            meta_ml_functions: Optional[List[Tuple[str, str, Callable[[np.array], float]]]] = None,
            cardinality_percentile: Optional[float] = None,
            radius_percentile: Optional[float] = None,
            lfd_percentile: Optional[float] = None,
            normalization_mode: Optional[str] = 'gaussian',
            speed_threshold: Optional[int] = 128,
    ):
        """ Creates and initializes a CHAODA object.

        :param metrics: A list of distance metrics to use for creating manifolds.
                        A metric must deterministically produce positive real numbers for each pair of instances.
                        A metric need not obey the triangle inequality.
                        Any such metrics allowed by scipy are allowed here, as are user-defined functions (so long as scipy accepts them).
        :param min_depth: The minimum depth clusters that can be selected for optimal graphs using meta-ml.
        :param max_depth: The max-depth of the cluster-trees in the manifolds.
        :param min_points: The minimum number of points in a cluster before it can be partitioned further.
        :param meta_ml_functions: A list of tuples of (metric-name, method-name, meta-ml-ranking-function).
        :param cardinality_percentile: The percentile of cluster cardinalities immediately under which clusters wil be selected for graphs.
        :param radius_percentile: The percentile of cluster radii immediately under which clusters wil be selected for graphs.
        :param lfd_percentile: The percentile of cluster local-fractal-dimensions immediately above which clusters wil be selected for graphs.
        :param normalization_mode: What normalization mode to use. Must be one of 'linear', 'gaussian', or 'sigmoid'.
        :param speed_threshold: number of clusters above which to skip the slow methods.
        """
        self.metrics: List[Metric] = ['euclidean', 'cityblock'] if metrics is None else metrics

        # Set criteria for building manifolds
        self.max_depth: int = max_depth
        self.min_points: int = min_points

        # Set meta-ml selection criteria
        self.min_depth: int = min_depth

        if meta_ml_functions is None:
            cardinality_percentile = 50 if cardinality_percentile is None else cardinality_percentile
            radius_percentile = 50 if radius_percentile is None else radius_percentile
            lfd_percentile = 25 if lfd_percentile is None else lfd_percentile
            criteria = [
                PropertyThreshold('cardinality', cardinality_percentile, 'below'),
                PropertyThreshold('radius', radius_percentile, 'below'),
                PropertyThreshold('lfd', lfd_percentile, 'above'),
            ]
            self._criteria = {
                metric: {method: criteria for method in self.method_names}
                for metric in self.metrics
            }
        else:
            metrics = {metric for metric, _, _ in meta_ml_functions}
            if not metrics <= set(self.metrics):
                raise ValueError(f'some meta-ml-functions reference metrics not being used by CHAODA. {set(self.metrics) - metrics}')
            methods = {method for _, method, _ in meta_ml_functions}
            if not methods <= set(self.method_names):
                raise ValueError(f'some meta-ml-functions reference methods not being used by CHAODA. {set(self.method_names) - methods}')

            self._criteria = {
                metric: {method: list() for method in self.method_names}
                for metric in self.metrics
            }
            for metric, method, function in meta_ml_functions:
                self._criteria[metric][method].append(MetaMLSelect(function, self.min_depth))

        self.speed_threshold: Optional[int] = speed_threshold

        if normalization_mode is not None:
            catch_normalization_mode(normalization_mode)
        self.normalization_mode: Optional[str] = normalization_mode

        # dict of name -> individual-method
        self._names: Dict[str, Callable[[Graph], ClusterScores]] = {
            'cluster_cardinality': self._cluster_cardinality,
            'component_cardinality': self._component_cardinality,
            'graph_neighborhood': self._graph_neighborhood,
            'parent_cardinality': self._parent_cardinality,
            'stationary_probabilities': self._stationary_probabilities,
            'vertex_degree': self._vertex_degree,
        }
        self.slow_methods: List[str] = [
            'component_cardinality',
            'graph_neighborhood',
            'stationary_probabilities',
        ]

        # Values to be set later

        # list of manifolds build during the fit method.
        self._manifolds: Optional[List[Manifold]] = None

        # These graphs are used with the individual-methods.
        #  This is a list of tuples os (method-name, graph).
        self._graphs: Optional[List[Tuple[str, List[Graph]]]] = None

        # A list of all outlier-scores arrays to be voted amongst.
        self._individual_scores: Optional[List[np.array]] = None

        # outlier-scores for each point in the fitted data, after voting.
        self._scores: Optional[np.array] = None

    @property
    def manifolds(self) -> List[Manifold]:
        """ Returns the list of manifolds being used for anomaly-detection. """
        return self._manifolds

    @property
    def scores(self) -> Optional[np.array]:
        """ Returns the scores for data on which the model was last fit. """
        if self._scores is None:
            logging.warning(f'Scores are currently empty. Please call the fit method.')
        return self._scores

    @property
    def individual_scores(self) -> Optional[List[np.array]]:
        if self._individual_scores is None:
            logging.warning(f'Individual-scores are currently empty. Please call the fit method.')
        return self._individual_scores

    def build_manifolds(self, data: np.array) -> List[Manifold]:
        """ Builds the list of manifolds for the class.

        :param data: numpy array of data where the rows are instances and the columns are features.
        :return: The list of manifolds.
        """
        criteria: List[Criterion] = [MaxDepth(self.max_depth), MinPoints(self.min_points)]
        self._manifolds = [Manifold(data, metric).build(*criteria) for metric in self.metrics]
        return self._manifolds

    def fit(self, data: np.array, *, voting: str = 'mean') -> 'CHAODA':
        """ Fits the anomaly detector to the data.

        :param data: numpy array of data where the rows are instances and the columns are features.
        :param voting: How to vote among scores.
        :return: the fitted CHAODA model.
        """
        self.build_manifolds(data)
        self._graphs = list()
        for manifold in self._manifolds:
            for method, meta_ml_criteria in self._criteria[manifold.metric].items():
                manifold.add_graphs(*meta_ml_criteria)
                for graph in manifold.graphs[-len(meta_ml_criteria):]:
                    if method in self.slow_methods and graph.cardinality > self.speed_threshold:
                        continue
                    self._graphs.append((method, graph))
        self._ensemble(voting)
        return self

    def predict(self, *, queries: np.array) -> np.array:
        # TODO: Handle seeing unseen data for live, online anomaly detection
        raise NotImplementedError

    def vote(self, voting: str = 'mean', replace: bool = True) -> np.array:
        """ Get ensemble scores with custom voting and normalization

        :param voting: voting mode to use. Must be one of VOTING_MODES.
        :param replace: whether to replace internal scores with new scores.
        :return: 1-d array of outlier scores for fitted data.
        """
        scores = self._vote(np.stack(self._individual_scores), voting)
        if replace:
            self._scores = scores
        return scores

    def cluster_cardinality(self, graph: Graph) -> Scores:
        """ Determines outlier scores for points by considering the relative cardinalities of the clusters in the graph.

        Points in clusters with relatively low cardinalities are the outliers.

        :param graph: Graph on which to calculate outlier scores.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._cluster_cardinality(graph))

    def component_cardinality(self, graph: Graph) -> Scores:
        """ Determines outlier scores by considering the relative cardinalities of the connected components of the graph.

        Points in components of relatively low cardinalities are the outliers

        :param graph: Graph on which to calculate outlier scores.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._component_cardinality(graph))

    def graph_neighborhood(self, graph: Graph, eccentricity_fraction: float = 0.25) -> Scores:
        """ Determines outlier scores by the considering the relative graph-neighborhood of clusters.

        Subsumed clusters are assigned the highest score of all subsuming clusters.
        Points in clusters with relatively small neighborhoods are the outliers.

        :param graph: Graph on which to calculate outlier scores.
        :param eccentricity_fraction: The fraction, in the (0, 1] range, of a cluster's eccentricity for which to compute the size of the neighborhood.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._graph_neighborhood(graph, eccentricity_fraction))

    def parent_cardinality(self, graph: Graph, weight: Callable[[int], float] = None) -> Scores:
        """ Determines outlier scores for points by considering ratios of cardinalities of parent-child clusters.

        The ratios are weighted by the child's depth in the tree, and are then accumulated for each point in each cluster in the graph.
        Points with relatively high accumulated ratios are the outliers.

        :param graph: Graph on which to calculate outlier scores.
        :param weight: A function for weighing the contribution of each depth.
                       Takes an integer depth and returns a weight.
                       Defaults to 1 / sqrt(depth).
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._parent_cardinality(graph, weight))

    def stationary_probabilities(self, graph: Graph, steps: int = 16) -> Scores:
        """ Compute the Outlier scores based on the convergence of a random walk with weighted edges on each component of the Graph.

        For each component on the graph, compute the convergent transition matrix for that graph.
        Clusters with low values in that matrix are the outliers.

        :param graph: The graph on which to compute outlier scores.
        :param steps: number of steps to wait for convergence.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._stationary_probabilities(graph, steps))

    def vertex_degree(self, graph: Graph) -> Scores:
        """ Compute the Outlier scores based on the degree of each vertex of the Graph.

        For each cluster in teh graph, its outlier score is inversely proportional to its degree in the graph.

        :param graph: The graph on which to compute outlier scores.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._vertex_degree(graph))

    def _ensemble(self, voting: str):
        """ Ensemble of individual methods.

        :param voting: How to vote among scores.
        """
        argpoints = self._manifolds[0].root.argpoints
        self._individual_scores = list()
        for name, graph in self._graphs:
            method = self._names[name]
            scores: Scores = self._score_points(method(graph))
            self._individual_scores.append(np.asarray([scores[i] for i in argpoints], dtype=float))

        # store scores for fitted data.
        self._scores = self.vote(voting)
        return

    # noinspection PyMethodMayBeStatic
    def _vote(self, scores: np.array, mode: str):
        """ Vote among all individual-scores for ensemble-score.

        :param scores: An array of shape (num_graphs, num_points) of scores to be voted among.
        :param mode: Voting mode to use. Must be one of VOTING_MODES
        """
        if mode not in VOTING_MODES:
            raise ValueError(f'voting mode {mode} is invalid. Must be one of {VOTING_MODES}.')

        if mode == 'mean':
            scores = np.mean(scores, axis=0)
        elif mode == 'product':
            scores = np.product(scores, axis=0)
        elif mode == 'median':
            scores = np.median(scores, axis=0)
        elif mode == 'min':
            scores = np.min(scores, axis=0)
        elif mode == 'max':
            scores = np.max(scores, axis=0)
        elif mode == 'p25':
            scores = np.percentile(scores, 25, axis=0)
        elif mode == 'p75':
            scores = np.percentile(scores, 75, axis=0)
        else:
            # TODO: Investigate other voting methods.
            pass

        return scores

    # noinspection PyMethodMayBeStatic
    def _score_points(self, scores: ClusterScores) -> Scores:
        """ Translate scores for clusters into scores for points. """
        scores = {point: float(score) for cluster, score in scores.items() for point in cluster.argpoints}
        for i in self._manifolds[0].root.argpoints:
            if i not in scores:
                # TODO: Rip off this band-aid and investigate source of missing points
                #  So far, this only triggers for exactly one point in exactly one dataset from our suite of datasets.
                #  No idea why.
                scores[i] = 0.5
        return scores

    def _normalize_scores(self, scores: Union[ClusterScores, Scores], high: bool) -> Union[ClusterScores, Scores]:
        """ Normalizes scores to lie in the [0, 1] range.

        :param scores: A dictionary of outlier rankings for clusters.
        :param high: True if high scores denote outliers, False if low scores denote outliers.

        :return: A dict of cluster -> normalized outlier score.
        """
        if self.normalization_mode is None:
            normalized_scores: List[float] = [(s if high is True else (1 - s)) for s in scores.values()]
        else:
            sign = 1 if high is True else -1
            normalized_scores: List[float] = list(normalize(
                values=sign * np.asarray([score for score in scores.values()], dtype=float),
                mode=self.normalization_mode,
            ))

        return {cluster: float(score) for cluster, score in zip(scores.keys(), normalized_scores)}

    def _cluster_cardinality(self, graph: Graph) -> ClusterScores:
        logging.info(f'with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        return self._normalize_scores({cluster: cluster.cardinality for cluster in graph.clusters}, False)

    def _component_cardinality(self, graph: Graph) -> ClusterScores:
        logging.info(f'with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        scores: Dict[Cluster, int] = {
            cluster: component.cardinality
            for component in graph.components
            for cluster in component.clusters
        }
        return self._normalize_scores(scores, False)

    # noinspection PyMethodMayBeStatic
    def _inherit_subsumed(self, subsumed_neighbors, scores):
        for master, subsumed in subsumed_neighbors.items():
            for cluster in subsumed:
                if cluster in scores:
                    scores[cluster] = max(scores[cluster], scores[master])
                else:
                    scores[cluster] = scores[master]
        return scores

    def _graph_neighborhood(self, graph: Graph, eccentricity_fraction: float = 0.25) -> ClusterScores:
        logging.info(f'Running method GN with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        if not (0 < eccentricity_fraction <= 1):
            raise ValueError(f'eccentricity fraction must be in the (0, 1] range. Got {eccentricity_fraction:.2f} instead.')

        pruned_graph, subsumed_neighbors = graph.pruned_graph

        def _neighborhood_size(start: Cluster, steps: int) -> int:
            """ Returns the number of clusters within 'steps' of 'start'. """
            visited: Set[Cluster] = set()
            frontier: Set[Cluster] = {start}
            for _ in range(steps):
                if frontier:
                    visited.update(frontier)
                    frontier = {
                        neighbor for cluster in frontier
                        for neighbor in pruned_graph.neighbors(cluster)
                        if neighbor not in visited
                    }
                else:
                    break
            return len(visited)

        scores: Dict[Cluster, int] = {
            cluster: _neighborhood_size(cluster, int(pruned_graph.eccentricity(cluster) * eccentricity_fraction))
            for cluster in pruned_graph.clusters
        }
        scores = self._inherit_subsumed(subsumed_neighbors, scores)
        return self._normalize_scores(scores, False)

    def _parent_cardinality(self, graph: Graph, weight: Callable[[int], float] = None) -> ClusterScores:
        logging.info(f'Running method PC with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')

        weight = (lambda d: 1 / (d ** 0.5)) if weight is None else weight
        scores: ClusterScores = {cluster: 0 for cluster in graph.clusters}
        for cluster in graph:
            ancestry = graph.manifold.ancestry(cluster)
            for i in range(1, len(ancestry)):
                scores[cluster] += (weight(i) * ancestry[i-1].cardinality / ancestry[i].cardinality)
        return self._normalize_scores(scores, True)

    def _stationary_probabilities(self, graph: Graph, steps: int = 16) -> ClusterScores:
        logging.info(f'Running method SP with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        scores: Dict[Cluster, float] = {cluster: -1 for cluster in graph.clusters}
        pruned_graph, subsumed_neighbors = graph.pruned_graph
        for component in pruned_graph.components:
            if component.cardinality > 1:
                clusters, matrix = component.as_matrix
                matrix = matrix / np.sum(matrix, axis=1)[:, None]

                for _ in range(steps):
                    # TODO: Go until convergence. For now, matrix ^ (2 ^ 16) ought to be enough.
                    matrix = np.linalg.matrix_power(matrix, 2)
                steady = matrix
                scores.update({cluster: score for cluster, score in zip(clusters, np.sum(steady, axis=0))})
            else:
                scores.update({cluster: 0 for cluster in component.clusters})
        scores = self._inherit_subsumed(subsumed_neighbors, scores)
        return self._normalize_scores(scores, False)

    def _vertex_degree(self, graph: Graph) -> ClusterScores:
        logging.info(f'Running method VD with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        scores: Dict[Cluster, float] = {cluster: graph.vertex_degree(cluster) for cluster in graph.clusters}
        return self._normalize_scores(scores, False)



from typing import Callable
from typing import List
from typing import Union

import numpy as np

Data = Union[np.memmap, np.ndarray]  # the entire dataset, TODO: Replace with custom DataLoader class
Datum = Union[int, np.ndarray]  # a point, or the index of a point, TODO: Use only np.array as Datum
Radius = Union[float, int, np.float64]
Vector = List[int]  # TODO: Rename to Indices of Argpoints
DistanceFunc = Callable[[Data, Data], Radius]
Metric = Union[str, DistanceFunc]  # TODO: Compile a list of metrics supported by scipy and by CLAM.



from typing import List
from typing import Tuple
from typing import Union

import numpy as np

__all__ = [
    "random",
    "bullseye",
    "line",
    "xor",
    "spiral_2d",
    "tori",
    "skewer",
]

np.random.seed(42)

Data = Union[np.ndarray, np.memmap]
Label = Union[np.ndarray, List[int]]


def random(n: int = 100, dimensions: int = 10) -> Tuple[Data, Label]:
    return np.random.randn(n, dimensions), np.zeros((n, dimensions), dtype=np.int)


def ring_data(n: int, radius: float, noise: float) -> np.ndarray:
    theta: np.ndarray = 2 * np.pi * np.random.rand(n)
    x: np.ndarray = radius * np.cos(theta) + noise * np.random.randn(n)
    y: np.ndarray = radius * np.sin(theta) + noise * np.random.randn(n)
    ring = np.stack([x, y], axis=1)
    return np.asarray(ring, dtype=np.float64)


def bullseye(n: int = 2_000, num_rings: int = 3, noise: float = 0.05) -> Tuple[np.ndarray, List[int]]:
    data: np.ndarray = np.ndarray(shape=(0, 2))
    labels: List[int] = list()
    for r in range(1, 2 * num_rings, 2):
        ring: np.ndarray = ring_data(n=n * r, radius=r, noise=noise)
        labels.extend([r for _ in range(n * r)])
        data = np.concatenate([data, ring], axis=0)
    return np.asarray(data, dtype=np.float64), labels


def line(n: int = 5_000, m: float = 1, c: float = 0., noise: float = 0.05) -> Tuple[np.ndarray, List[int]]:
    x = np.random.rand(n)
    y = m * x + c
    data = np.asarray((x, y)).T
    data = data + np.random.rand(*data.shape) * noise
    labels = np.ones_like(x.T)
    return np.asarray(data, dtype=np.float64), list(labels)


def xor(n: int = 5_000) -> Tuple[np.ndarray, List[int]]:
    data = np.random.rand(n, 2)
    labels = [int((x > 0.5) != (y > 0.5)) for x, y, in data]
    return np.asarray(data, dtype=np.float64), labels


def spiral_2d(n: int = 5_000, noise: float = 0.1) -> Tuple[np.ndarray, List[int]]:
    theta = np.sqrt(np.random.rand(n)) * 2 * np.pi

    r_a = 2 * theta + np.pi
    data_a = np.array([np.cos(theta) * r_a, np.sin(theta) * r_a]).T
    x_a = data_a + np.random.randn(n, 2) * noise

    r_b = -2 * theta - np.pi
    data_b = np.array([np.cos(theta) * r_b, np.sin(theta) * r_b]).T
    x_b = data_b + np.random.randn(n, 2) * noise

    data = np.concatenate([x_a, x_b]) / 5
    labels = list(np.concatenate([np.zeros(len(x_a)), np.ones(len(x_a))]))
    return np.asarray(data, dtype=np.float64), labels


def generate_torus(n: int, r_torus: float, noise: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r_tube: float = r_torus / 5
    u, v = np.random.rand(n), np.random.rand(n)
    u, v = u * 2 * np.pi, v * 2 * np.pi
    x = (r_torus + r_tube * np.cos(v)) * np.cos(u) + (np.random.randn(n) * noise)
    y = (r_torus + r_tube * np.cos(v)) * np.sin(u) + (np.random.randn(n) * noise)
    z = r_tube * np.sin(v) + (np.random.randn(n) * noise)
    return x, y, z


def tori(n: int = 10_000, noise: float = 0.015, r_torus: float = 1.) -> Tuple[np.ndarray, List[int]]:
    x, y, z = generate_torus(n=n // 2, r_torus=r_torus, noise=noise)
    torus_1 = np.stack([x - r_torus, y, z], axis=1)
    labels = [0 for _ in x]

    x, y, z = generate_torus(n=n // 2, r_torus=r_torus, noise=noise)
    torus_2 = np.stack([x, z, y], axis=1)
    labels.extend([1 for _ in x])

    data = np.concatenate([torus_1, torus_2], axis=0)
    return np.asarray(data, dtype=np.float64), labels


def spiral_3d(n: int, radius: float, height: float, num_turns: int, noise: float) -> np.ndarray:
    theta: np.ndarray = 2 * np.pi * np.random.rand(n) * num_turns
    x: np.ndarray = radius * np.cos(theta) + noise * np.random.randn(n)
    y: np.ndarray = radius * np.sin(theta) + noise * np.random.randn(n)
    z: np.ndarray = height * theta / (2 * np.pi * num_turns) + noise * np.random.randn(n)
    return np.stack([x, y, z], axis=1)


def line_3d(n: int, height: float, noise: float) -> np.ndarray:
    x: np.ndarray = noise * np.random.randn(n)
    y: np.ndarray = noise * np.random.randn(n)
    z: np.ndarray = height * np.random.rand(n)
    return np.stack([x, y, z], axis=1)


def skewer(
        n: int = 10_000,
        radius: float = 3.,
        height: float = 6.,
        num_turns: int = 2,
        noise: float = 0.05
) -> Tuple[np.ndarray, List[int]]:
    spiral_points, line_points = 5 * n // 6, n // 6
    spiral_data = spiral_3d(spiral_points, radius, height, num_turns, noise)
    labels = [0 for _ in range(spiral_data.shape[0])]

    line_data = line_3d(line_points, height, noise)
    labels.extend([1 for _ in range(line_data.shape[0])])

    data = np.concatenate([spiral_data, line_data], axis=0)
    return np.asarray(data, dtype=np.float64), labels



from . import criterion
from . import datasets
from . import types
from . import utils
from .chaoda import CHAODA
from .manifold import Cluster
from .manifold import Edge
from .manifold import Graph
from .manifold import Manifold
from .search import Search



from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np

import pyclam.criterion as criterion
from pyclam.manifold import Cluster
from pyclam.manifold import EPSILON
from pyclam.manifold import Manifold
from pyclam.types import Data
from pyclam.types import Datum
from pyclam.types import Metric

ClusterResults = Dict[Cluster, float]
Results = Dict[int, float]


class Search:
    def __init__(self, data: Data, metric: Metric):
        self.data: Data = data
        self.metric: Metric = metric
        self.manifold: Manifold = Manifold(self.data, self.metric)
        self.distance = self.manifold.distance
        self.root: Cluster = self.manifold.root

    @property
    def depth(self) -> int:
        """ returns the depth of the search tree. """
        return self.manifold.depth

    def build(self, *, max_depth: Optional[int] = None) -> 'Search':
        """ Builds the search tree upto leaves, or an optional maximum depth.

        This method can be called repeatedly, with higher depth values, to further increase the depth of the tree.

        :param max_depth: optional maximum depth of search tree
        :return: the modified Search object.
        """
        if max_depth is None:
            self.manifold.build(criterion.Layer(-1))
        elif max_depth < 1:
            raise ValueError(f'Expected a positive integer for max_depth. Got {max_depth} instead.')
        elif max_depth > self.depth:
            self.manifold.build_tree(criterion.MaxDepth(max_depth), criterion.Layer(-1))
        return self

    def rnn(self, query: Datum, radius: float, *, max_depth: Optional[int] = None) -> Results:
        """ Performs rho-nearest neighbors search around query with given radius.

        :param query: point around which to search.
        :param radius: distance from query within which te return all hits.
        :param max_depth: optional maximum search depth.
        :return: dictionary of index-of-hit -> distance-to-hit.
        """
        candidate_clusters: ClusterResults = self.tree_search(query, radius, self.root, max_depth=max_depth)
        return self.leaf_search(query, radius, candidate_clusters)

    def rnn_points(self, query: Datum, radius: float, *, max_depth: Optional[int] = None) -> Data:
        """ Same as rnn, but returns the raw points rather than a dictionary of indices and distances. """
        results = [(d, p) for p, d in self.rnn(query, radius, max_depth=max_depth).items()]
        return self.data[[p for _, p in sorted(results)]]

    def knn(self, query: Datum, k: int, *, max_depth: Optional[int] = None) -> Results:
        """ Performs k-nearest neighbors search around query.

        :param query: point around which to look.
        :param k: number of closest neighbors to look for.
        :param max_depth: optional maximum search depth.
        :return: dictionary of index-of-neighbor -> distance-to-neighbor.
        """
        if k < 1:
            raise ValueError(f'k must be a positive integer. Got {k}')

        radius: float = (k / self.root.cardinality) * self.root.radius
        assert radius > 0, f'expected a positive value for radius. Got {radius:.2e} instead.'

        hits: Results = self.rnn(query, radius, max_depth=max_depth)
        while len(hits) == 0:  # make sure to have non-zero hits
            radius *= 2
            hits = self.rnn(query, radius, max_depth=max_depth)

        while len(hits) < k:  # make sure to have more at least k hits
            # calculate lfd in ball around query
            if len(hits) == 1:
                lfd = 1
            else:
                half_count = len([point for point, distance in hits.items() if distance <= radius / 2])
                lfd = 1 if half_count == 0 else np.log2(len(hits) / half_count)

            # increase radius as guided by lfd
            factor = (k / len(hits)) ** (1 / (lfd + EPSILON))
            assert factor > 1, f'expected factor to be greater than 1. Got {factor:.2e} instead.'
            radius *= factor

            # rerun rnn search
            hits = self.rnn(query, radius, max_depth=max_depth)

        # sort hits in non-decreasing order of distance to query
        results: List[Tuple[float, int]] = list(sorted([(distance, point) for point, distance in hits.items()]))
        return {point: distance for distance, point in results[:k]}

    def knn_points(self, query: Datum, k: int, *, max_depth: Optional[int] = None) -> Data:
        """ Same as knn, but returns the raw points rather than a dictionary of indices and distances. """
        results = [(d, p) for p, d in self.knn(query, k, max_depth=max_depth).items()]
        return self.data[[p for _, p in sorted(results)]]

    def _check_shape(self, points: np.ndarray) -> None:
        if len(points.shape) != len(self.data.shape):
            raise ValueError(f'wrong number of dimensions in shape. Expected {len(self.data.shape)}, but got {len(points.shape)}')
        elif points.shape[1:] != self.data.shape[1:]:
            raise ValueError(f'wrong dimensionality of points. Expected {self.data.shape[1:]}, but got {points.shape[1:]}')
        else:
            return

    def _parse_query(self, query: Datum) -> np.ndarray:
        if isinstance(query, int):
            query: np.ndarray = np.asarray([self.data[query]])
        query = np.expand_dims(query, 0)
        self._check_shape(query)
        return query

    def tree_search(self, query: Datum, radius: float, start: Cluster, *, max_depth: Optional[int] = None) -> ClusterResults:
        """ Performs tree-search for the query, starting at the given cluster.

        Consider the sphere around 'query' with the given 'radius'.
        This method performs a breadth-first search over the tree and, at each iteration,
        discards the clusters that do not have overlap with this sphere.
        Once the search terminates, the remaining clusters are returned.

        :param query: point around which to search.
        :param radius: distance from point within which to look.
        :param start: starting cluster for search.
        :param max_depth: optional maximum search depth starting at Cluster.
        :return: dictionary of cluster -> distance between query and cluster center
        """
        return self.tree_search_history(query, radius, start, max_depth=max_depth)[1]

    def tree_search_history(self, query: Datum, radius: float, start: Cluster, *, max_depth: Optional[int] = None) -> Tuple[ClusterResults, ClusterResults]:
        """ Same as tree-search, except that it also returns the history of candidate clusters at each depth. """
        if max_depth is None:
            max_depth = self.manifold.depth
        elif max_depth < 0:
            raise ValueError(f'max_depth must be a non-negative integer. Got {max_depth} instead.')
        else:
            max_depth = min(self.manifold.depth, start.depth + max_depth)

        query = self._parse_query(query)
        candidates: ClusterResults = {start: -1}
        history: ClusterResults = dict()
        hits: ClusterResults = dict()
        for depth in range(start.depth, max_depth + 1):
            if len(candidates) == 0:
                break
            # find distances to centers of candidate clusters
            clusters: List[Cluster] = list(candidates.keys())
            centers: np.ndarray = np.asarray([cluster.center for cluster in clusters])
            self._check_shape(centers)
            distances: List[float] = list(self.distance(query, centers)[0])

            candidates = {  # filter out clusters that are too far away
                cluster: distance for cluster, distance in zip(clusters, distances)
                if distance <= (radius + cluster.radius)
            }
            subsumed: Set[Cluster] = {  # set of clusters that lie completely within radius of query
                cluster for cluster, distance in candidates.items()
                if (distance + cluster.radius) <= radius
            }
            hits.update({  # add subsumed clusters to hits, singletons should already be included here
                cluster: distance
                for cluster, distance in candidates.items()
                if (cluster in subsumed) or (len(cluster.children) == 0)
            })

            history.update(candidates)  # add candidates from this depth to history

            candidates = {  # expand search to children
                child: -1
                for cluster in candidates.keys()
                for child in cluster.children
                if cluster not in hits
            }

        return history, hits

    def leaf_search(self, query: Datum, radius: float, clusters: ClusterResults) -> Results:
        """ Performs leaf search for query on given clusters.

        :param query: point around which to look.
        :param radius: distance from point within which to look.
        :param clusters: candidate clusters which need to be exhaustively searched.
        :return: dictionary of index-of-hit -> distance-to-hit.
        """
        query = self._parse_query(query)

        # initialize hits
        hits: Results = dict()

        # get indices of all points in candidate clusters
        argpoints: List[int] = [point for cluster in clusters for point in cluster.argpoints]

        # get distances from query to candidate points
        points: np.ndarray = self.data[argpoints]
        self._check_shape(points)
        distances: List[float] = list(self.distance(query, points)[0])

        # update hits with points within radius of query
        hits.update({point: distance for point, distance in zip(argpoints, distances) if distance <= radius})

        return hits



""" Some common functions and constants for all of CLAM.
"""
SUBSAMPLE_LIMIT = 100
BATCH_SIZE = 10_000
EPSILON = 1e-8


def catch_normalization_mode(mode: str) -> None:
    from typing import List

    """ Make sure that the normalization mode is allowed. """
    modes: List[str] = ['linear', 'gaussian', 'sigmoid']
    if mode not in modes:
        raise ValueError(f'Normalization method {mode} is undefined. Must by one of {modes}.')
    else:
        return


def normalize(values, mode: str):
    """ Normalizes each column in values into a [0, 1] range.

    :param values: A 1-d or 2-d array of values to normalize.
    :param mode: Normalization mode to use. Must be one of 'linear', 'gaussian', or 'sigmoid'.
    :return: array of normalized values.
    """
    import numpy as np

    squeeze = False
    if len(values.shape) == 1:
        squeeze = True
        values = np.expand_dims(values, axis=1)

    if mode == 'linear':
        min_v, max_v = np.min(values, axis=0), np.max(values, axis=0)
        for i in range(values.shape[1]):
            if min_v[i] == max_v[i]:
                max_v[i] += 1
                values[:, i] = min_v[i] + 0.5
        values = (values - min_v) / (max_v - min_v)
    else:
        mu = np.mean(values, axis=0)
        sigma = np.std(values, axis=0)

        for i in range(values.shape[1]):
            if sigma[i] < EPSILON:
                values[:, i] = 0.5
            else:
                if mode == 'gaussian':
                    from scipy.special import erf

                    values[:, i] = (1 + erf((values[:, i] - mu[i]) / (sigma[i] * np.sqrt(2)))) / 2
                else:
                    values[:, i] = 1 / (1 + np.exp(-(values[:, i] - mu[i]) / sigma[i]))

    values = values.clip(EPSILON, 1)
    if squeeze:
        values = np.squeeze(values)
    return values
